;		Dynamic scoping with shift/reset
;
; Delimited continuation operators implicitly have `dynamic scope'.  If
; we have dynamically scoped variables (which truly follow the stack,
; even when some parts of it are skipped due to captured and invoked
; first-class continuations), and call/cc, we can implement delimited
; continuation operators. See the implementation of letSubCont via
; `holes', which are dynamically scoped.
;
; The present code shows the reverse operation: implementing dynamic scoping
; via shift/reset. Binding and dereference of fluid variables are
; *macro*-expressible in terms of shift/reset; no mutable state nor
; undelimited continuations are used.
;
; Previously,
; Dynamically-scoped (aka fluid) variables must be declared first,
; using the form `define-fluid-var'.
; One can define a form `let-fluid-var' to define fluid-vars in a local
; scope. One can use `let-fluid-var' to bind a fluid variable to a value.
; To dereference a fluid variable (that is, to obtain its value),
; we simply do `(var)' -- the variable is actually a thunk, so we just
; invoke that thunk.
;
; In the new version, the naming of dynamic variables is manifest:
; we can use any symbol (or number, etc. - anything that can be equal?
; comparable. We use equal? for purity. We will be using symbols anyway,
; in which case equal? is equivalent to eq?).
; Dynamic variables do not have to be declared.
; we use the function (fluid-get name) to dereference the variable `name'.
;
; The technique is almost the same as the one used to emulate `control'
; and other dynamic delimited continuation operators via shift/reset.
;
; The comparison with the corresponding Haskell code shows that multi-prompt
; cupto etc. operators are macro-expressible in terms of a single-prompt
; shift/reset.

; On the issue of naming
;
; If we have more than one dynamically-bound variable, we need some way
; of telling one from the other. When referring to the value associated
; with a variable, we need to somehow identify which variable we are
; talking about. This is an interesting problem for dynamic scoping,
; because such variables are useful precisely when the variable
; reference is _not_ in the scope of its binding form. There are two
; approaches: one is to choose some `manifest' means of identification
; -- like a number, a string, etc. Haskell implicit parameters takes
; this approach: implicit parameters are identified solely by their
; spelling. No `declaration' of an implicit parameter is needed. The
; current implementation of dynamically scoped parameters essentially
; takes this route too. I can indeed easily get rid of the declaration
; of dynamically scoped variables and identify the variables by strings,
; for example, or by integers.
;
; The other approach is to ``create'' dynamically scoped variables as
; first-class objects. Such objects should be bound to some lexical
; variables. SRFI-39 `Parameter objects' takes this approach:
;         http://srfi.schemers.org/srfi-39/srfi-39.html
;
; The issue of the identification still remains. In SRFI-39, it is
; solved by using eq? on procedural values, which is quite ugly. 
;
; It seems we can implement dynamically scoped variables that way too --
; and *avoid* the use of eq?, (gensym), or explicit name tags. The
; technique is simple: the identification tag will be chosen by the
; system -- on an explicit `request' delivered by `shift'. The downside
; is that dynamically scoped variables cannot be created on top-level.
; The whole program has to be wrapped in a prompt (which actually
; chooses the identification labels).


; $Id: dynamic-scope-shift.scm,v 1.5 2005/09/15 05:59:37 oleg Exp $


; To get shift/reset on Scheme48, do
; ,open escapes signals
; ,load /usr/local/lib/scheme48/misc/shift-reset.scm


; data H = H a b | HV v | HUser a b
; Here is an implementation of the H datatype, as an indiscriminated 
; Scheme union
; See delim-cont.scm for a pure realization, as pure lambda terms.

(define H-tag     (list 'H-tag))
(define HUser-tag (list 'HUser-tag))
(define HV-tag    (list 'HV-tag))

; Constructors
(define (H a b) (cons H-tag (cons a b)))
(define (HUser a b) (cons HUser-tag (cons a b)))

; we specifically tag HV alternative to make sure our tagging is correct
(define (HV v) (cons HV-tag v))

; (define-syntax HV			; just the identity
;   (syntax-rules ()
;     ((HV v) v)))

; Deconstructor
(define-syntax case-H
  (syntax-rules ()
    ((case-H e
       ((f x) on-h)
       ((f1 x1) on-h-user)
       (v on-hv))
      (let ((val e))
	(cond
	  ((and (pair? val) (eq? (car val) H-tag))
	    (let ((f (cadr val)) (x (cddr val))) on-h))
	  ((and (pair? val) (eq? (car val) HUser-tag))
	    (let ((f1 (cadr val)) (x1 (cddr val))) on-h-user))
	  ((and (pair? val) (eq? (car val) HV-tag))
	    (let ((v (cdr val))) on-hv))
	  (else
	    (error "missing tag")))))))

; User-level shift/reset
; See greset/gshift and implementation of shift/reset via them in the TR

(define-syntax ureset
  (syntax-rules ()
    ((ureset e)
      (hr-stop (reset (HV e))))))

(define-syntax ushift
  (syntax-rules ()
    ((ushift f e)
      (shift f* (HUser (lambda (x) (hs-stop (f* x))) (lambda (f) e))))))


(define (hr-stop v)
  (display "hr-stop") (newline)
  (case-H v
    ; on-h: propagate up, compose with the fn
    ((v k) (begin
	     ;(display "vreq: ") (display v) (newline)
	     (let ((r (shift f (H v f))))
	       ;(display "got back: ") (display r) (newline)
	       (hr-stop (k r)))))
    ;(shift f (H v (lambda (x) (k (f x))))))
    ; on-h-user
    ((f x) (ureset (x f)))
    ; on-hv
    (v v)))

(define (hs-stop v)
  (display "hs-stop") (newline)
  (case-H v
    ; on-h: propagate up, compose with the fn
    ((v k) (begin
	      ;(display "vreq: ") (display v) (newline)
	     (let ((r (shift f (H v f))))
	       ;(display "got back: ") (display r) (newline)
	       (hs-stop (k r)))))
    ;(shift f (H v (lambda (x) (k (f x))))))
    ; on-h-user
    ((f x) (ureset (x f)))
    ; on-hv
    (v v)))

; That is, to request the value of a fluid variable, we `throw'
; a request upstairs. `H v k' is the request, in the H envelope.
; The name of the variable is the `addressee' of the request, and
; 'k' is the `return address'.
; (define-syntax define-fluid-var
;   (syntax-rules ()
;     ((define-fluid-var var)
;       (define var (lambda () (shift f (H 'var f)))))))

(define (fluid-get name)
  (shift f (H name f)))

(define-syntax fluid-let
  (syntax-rules ()
    ((fluid-let ((var init)) body)
      (let ((val init))
	(let loop ((result (reset (HV body))))
	  (display "in loop of: ") (display 'var) (display val)(newline)
	  ; check to see if we've got a request in the H envelope
	  (case-H result
	    ((requested-var k) ; request for a fluid-var's value
	      (if (equal? requested-var 'var) ; our var?
		(loop (k val))		; give it to k
		                        ; Otherwise, refer to upstairs
					; Note the implicit composition of
					; g and k
		(loop (k (shift g (H requested-var g))))))
	    ; HUser shift: propagate up
	    ; see function hs-prop in TR
	    ((f x)
	      (begin (display "HUser in fluid-let") (newline)
	      (shift g
		(HUser (lambda (y) (hs-stop (g (loop (reset (HV (f y))))))) x))))
	    (v v)))))))


; (define (hs-prop v)
;   (display "hs-stop") (newline)
;   (case-H v
;     ; on-h
;     ((v k)
;       (shift f
; 	(H v (lambda (y) (k (f y))))))
;     ; on-h-user
;     ((f x)
;       (shift g
; 	(HUser (lambda (y) (hs-prop (g (f y)))) x)))
;     ; on-hv
;     (v v)))

; Tests


; Declare all the variables first
; (define-fluid-var a)
; (define-fluid-var b)
; (define-fluid-var c)

(define (print-it title)
  (for-each display
    (list
      title #\newline
      "var a: " (fluid-get 'a) #\newline
      "var b: " (fluid-get 'b) #\newline
      "var c: " (fluid-get 'c) #\newline
      #\newline)))

; First we test that fluid variables are first-class, and they can be
; passed `by name'
(define (test1 var)
  (display "passed var's val: ") (display (fluid-get var)) (newline)
  (fluid-let ((b 120))
    (begin
      (print-it "In test1")
      (let ((thunk (test2)))
	(thunk)))))

(define (test2)
  (fluid-let ((a 110))
    (fluid-let ((c 130))
      (begin
	(print-it "In test2")
	; return a `closure' -- but fluid vars are NOT captured
	(lambda () (print-it "``closure'' from test2"))))))


; (define (test3)
;   (print-it "In test3")
;   (let ((re-enter #f))
;     (let ((r
; 	    (fluid-let ((a 310))
; 	      ((call-with-current-continuation
; 		 (lambda (k) (set! re-enter k)
; 		   (lambda () (print-it "inside fluid-let") #t)))))))
;       (print-it "outside fluid-let")
;       (if r (re-enter (lambda () (print-it "inside again") #f)))
;       (print-it "outside again"))))


(define (test3-pure)
  (display "In test3")
  (let ((r
	  (ureset
	    (begin
	      ((ushift f 
		 (begin
		   (f (lambda () (display "inside fluid-let") f)))))))))
      (display "outside fluid-let")
      (if r (r (lambda () (display "inside again") #f)))
      (display "outside again")))

(define (test3)
  (print-it "In test3")
  (let ((r
	  (ureset
	    (fluid-let ((a 310))
	      ((ushift f 
		 (begin
		   (f (lambda () (print-it "inside fluid-let") f)))))))))
      (print-it "outside fluid-let")
      (display r)
      (r (lambda () (print-it "inside again")))
      (print-it "outside again")))		    

(define (test4)
  (print-it "In test4")
  (let ((r
	 (fluid-let ((a 310))
	  (ureset
	      ((ushift f 
		 (begin
		   (f (lambda () (print-it "inside fluid-let") f)))))))))
      (print-it "outside fluid-let")
      (display r)
      (r (lambda () (print-it "inside again")))
      (print-it "outside again")))		    

(define (init)
  (fluid-let ((a 10))
    (fluid-let ((b 20))
      (fluid-let ((c 30))
	(begin
	  (print-it "In init")
	  (test1 'c)
	  (test2)
	  (test3)
	  (test4)
	  (print-it "In init, at the end")
	  )))))

