;	         A counter-example
; to show that a simple implementation of dynamic variables via
; SML-like exceptions and call/cc breaks down in the presence of
; delimited continuations (shift/reset)
;
; This is R5RS code, tested with Petite Chez Scheme, SCM, and Scheme48.


;------------------------------------------------------------------------
; First, some preliminaries: ML-like exceptions handling for Scheme
;
; The following code is taken verbatim from SRFI-34
; Exception Handling for Programs
; http://srfi.schemers.org/srfi-34/srfi-34.html

(define *current-exception-handlers*
  (list (lambda (condition)
          (error "unhandled exception" condition))))

(define (with-exception-handler handler thunk)
  (with-exception-handlers (cons handler *current-exception-handlers*)
                           thunk))

(define (with-exception-handlers new-handlers thunk)
  (let ((previous-handlers *current-exception-handlers*))
    (dynamic-wind
      (lambda ()
        (set! *current-exception-handlers* new-handlers))
      thunk
      (lambda ()
        (set! *current-exception-handlers* previous-handlers)))))

(define (raise obj)
  (let ((handlers *current-exception-handlers*))
    (with-exception-handlers (cdr handlers)
      (lambda ()
        ((car handlers) obj)
        (error "handler returned"
               (car handlers)
               obj)))))

(define-syntax guard
  (syntax-rules ()
    ((guard (var clause ...) e1 e2 ...)
     ((call-with-current-continuation
       (lambda (guard-k)
         (with-exception-handler
          (lambda (condition)
            ((call-with-current-continuation
               (lambda (handler-k)
                 (guard-k
                  (lambda ()
                    (let ((var condition))      ; clauses may SET! var
                      (guard-aux (handler-k (lambda ()
                                              (raise condition)))
                                 clause ...))))))))
          (lambda ()
            (call-with-values
             (lambda () e1 e2 ...)
             (lambda args
               (guard-k (lambda ()
                          (apply values args)))))))))))))

(define-syntax guard-aux
  (syntax-rules (else =>)
    ((guard-aux reraise (else result1 result2 ...))
     (begin result1 result2 ...))
    ((guard-aux reraise (test => result))
     (let ((temp test))
       (if temp
           (result temp)
           reraise)))
    ((guard-aux reraise (test => result) clause1 clause2 ...)
     (let ((temp test))
       (if temp
           (result temp)
           (guard-aux reraise clause1 clause2 ...))))
    ((guard-aux reraise (test))
     test)
    ((guard-aux reraise (test) clause1 clause2 ...)
     (let ((temp test))
       (if temp
           temp
           (guard-aux reraise clause1 clause2 ...))))
    ((guard-aux reraise (test result1 result2 ...))
     (if test
         (begin result1 result2 ...)
         reraise))
    ((guard-aux reraise (test result1 result2 ...) clause1 clause2 ...)
     (if test
         (begin result1 result2 ...)
         (guard-aux reraise clause1 clause2 ...)))))



;------------------------------------------------------------------------
; The second part of preliminaries:
; Danvy/Filinski implementation of shift/reset, the Christmas present
; to Scheme48 authors.

(define-syntax reset
  (syntax-rules ()
    ((_ ?e ?f ...) (*reset (lambda () ?e ?f ...)))))

(define-syntax shift
  (syntax-rules ()
    ((_ ?k ?e ?f ...) (*shift (lambda (?k) ?e ?f ...)))))

(define *meta-continuation*
  (lambda (v)
    (error "You forgot the top-level reset...")))

(define *abort
  (lambda (thunk)
    ;(with-continuation null-continuation ;JAR hack
      ;(lambda ()
	(let ((val (thunk)))
	  (*meta-continuation* val))));))

(define *reset
  (lambda (thunk)
    (let ((mc *meta-continuation*))
      (call-with-current-continuation
	(lambda (k)
	  (begin
	    (set! *meta-continuation*
		  (lambda (v)
		    (set! *meta-continuation* mc)
		    (k v)))
	    (*abort thunk)))))))

(define *shift
  (lambda (f)
    (call-with-current-continuation
      (lambda (k)
	(*abort (lambda ()
		  (f (lambda (v)
		       (reset (k v))))))))))


;------------------------------------------------------------------------
; The following is the implementation of dynamic variables via
; exceptions and call/cc, suggested by POPL Reviewer 3
;
; For simplicity, we use only one dynvar, so we don't bother with
; naming it explicitly. For compatibility with the SML code (see *.sml
; files in this directory) we don't make dlet a macro -- hence dlet
; takes a thunk as the second argument.

; dereference the (implicit) dynamic variable
(define (dref)
  (call-with-current-continuation
    (lambda (k)
      (raise k))))

; Evaluate THUNK in the dynamic environment extended with the binding of the
; (implicit) dynvar to V
(define (dlet v thunk)
  (guard (k (else (k v)))
    (thunk)))

; The first test
(define (test1)
  (dlet 1
    (lambda ()
      (list (dref) (dref)
      (dlet 2 
	(lambda ()
	  (list (dref) (dref))))))))

(display (test1)) (newline)
; (1 1 (2 2))


; The counter-example itself

; A sample function to print the current dynamic environment, twice
; (just for the heck of it).
(define (tf title)
  (display title) (newline)
  (display (dref)) (newline)
  (display (dref)) (newline))

; The first test: we capture a delimited continuation that 
; DOES include `dlet 2'
(define (test2)
  (newline) (display "test2") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
	      (reset
		(dlet 2 
		  (lambda ()
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test2)

; The second test: we now capture a delimited continuation that does NOT
; include `dlet 2'
(define (test3)
  (newline) (display "test3") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
		(dlet 2 
		  (lambda ()
	      (reset
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test3)

; We observe, test2 and test3 print the same results. But that is wrong!
; In test3, the dynamic environment in the case "inside again"
; (that is, when the captured continuation is reinstated)
; must be 1 rather than 2 because `dlet 2' is NOT included within the
; captured continuation.
