(*
	         A counter-example
 to show that a simple implementation of dynamic variables via
 SML-like exceptions and call/cc breaks down in the presence of
 delimited continuations (shift/reset)

 The corresponding Scheme code: dynvar-via-exc.scm
*)

(*
 The following is the implementation of dynamic variables via
 exceptions and call/cc, suggested by POPL Reviewer 3.
 

 For simplicity, we use only one dynvar, so we don't bother with
 naming it explicitly. We also adjust the reviewer's code to use the 
 Escape structure from Filinski thesis (see escape.sml in this directory).
 It just gives call/cc a bit more pleasant type, and makes the code
 consistent with Filinski's implementation of shift (the Control structure).

*)

structure CEX = struct


(* The exception that corresponds to the implicit dynamic variable *)
exception Dyn of (int -> Escape.void)

(* dereference the (implicit) dynamic variable *)
fun dref () = Escape.escape (fn k => raise (Dyn k))

(* Evaluate thunk TH in the dynamic environment extended with 
   the binding of the (implicit) dynvar to V
*)
fun dlet v th = th () handle Dyn k => Escape.coerce (k v)

(* The first test, to check that binding and referencing really works *)
fun test1 () = 
  dlet 1 (fn () => [dref (), dref ()] @ (dlet 2 (fn () => [dref (), dref ()])))

(* The counter-example itself *)

(* A sample function to print the current dynamic environment, twice
  (just for the heck of it).
*)

fun tf title = 
  (print title;
   print "\n";
   print (Int.toString (dref ()));
   print "\n";
   print (Int.toString (dref ()));
   print "\n")

datatype rect = 
    J of (rect -> rect)
  | R of (string -> unit) 

structure CI = Control (type ans = rect)

(* The first test: we capture a delimited continuation that 
  DOES include `dlet 2'
*)

fun test2 () = 
    dlet 1 (fn () => 
      let val (J f) = 
                    CI.reset (fn () => 
                    dlet 2 (fn () => 
                    let val R f1 = CI.shift (fn f => (tf "inside shift"; J f))
                    in f1 "inside again"; R (fn x => ())
                    end))
	in (tf "outside"; f (R tf); tf "outside again")
      end)

(*
  The second test: we now capture a delimited continuation that does NOT
  include `dlet 2'
*)

fun test3 () = 
    dlet 1 (fn () => 
      let val (J f) = 
                    dlet 2 (fn () => 
                    CI.reset (fn () => 
                    let val R f1 = CI.shift (fn f => (tf "inside shift"; J f))
                    in f1 "inside again"; R (fn x => ())
                    end))
	in (tf "outside"; f (R tf); tf "outside again")
      end)

(*
 We observe, test2 and test3 print the same results. But that is wrong!
 In test3, the dynamic environment in the case "inside again"
 (that is, when the captured continuation is reinstated)
 must be 1 rather than 2 because `dlet 2' is NOT included within the
 captured continuation.
*)
end
